# SPDX-FileCopyrightText: 2022-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import os
import subprocess
import sys


def create_file(path):
    with open(path, 'a'):
        os.utime(path, None)


def read_file(path):
    with open(path, 'r') as f:
        return f.read()


def write_file(path, contents):
    with open(path, 'w') as f:
        f.write(contents)


def build_project(*args):
    _run_command(sys.executable, '-m', 'hatchling', 'build', *args)


def git(*args):
    _run_command('git', *args)


def _run_command(*command):
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout, _ = process.communicate()

    if process.returncode:  # no cov
        raise Exception(stdout.decode('utf-8'))



# SPDX-FileCopyrightText: 2022-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import errno
import os
import shutil
import stat
import tempfile
from contextlib import contextmanager

import pytest

from .utils import create_file, git, write_file


def handle_remove_readonly(func, path, exc):  # no cov
    # PermissionError: [WinError 5] Access is denied: '...\\.git\\...'
    if func in (os.rmdir, os.remove, os.unlink) and exc[1].errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        func(path)
    else:
        raise


@pytest.fixture
def temp_dir():
    directory = tempfile.mkdtemp()
    try:
        directory = os.path.realpath(directory)
        yield directory
    finally:
        shutil.rmtree(directory, ignore_errors=False, onerror=handle_remove_readonly)


@contextmanager
def create_project(directory, metadata, setup_vcs=True):
    project_dir = os.path.join(directory, 'my-app')
    os.mkdir(project_dir)

    gitignore_file = os.path.join(project_dir, '.gitignore')
    write_file(gitignore_file, '/my_app/version.py')

    project_file = os.path.join(project_dir, 'pyproject.toml')
    write_file(project_file, metadata)

    package_dir = os.path.join(project_dir, 'my_app')
    os.mkdir(package_dir)

    create_file(os.path.join(package_dir, '__init__.py'))
    create_file(os.path.join(package_dir, 'foo.py'))
    create_file(os.path.join(package_dir, 'bar.py'))
    create_file(os.path.join(package_dir, 'baz.py'))

    origin = os.getcwd()
    os.chdir(project_dir)
    try:
        if setup_vcs:
            git('init')
            git('config', '--local', 'user.name', 'foo')
            git('config', '--local', 'user.email', 'foo@bar.baz')
            git('add', '.')
            git('commit', '-m', 'test')
            git('tag', '1.2.3')

        yield project_dir
    finally:
        os.chdir(origin)


@pytest.fixture
def new_project_basic(temp_dir):
    with create_project(
        temp_dir,
        """\
[build-system]
requires = ["hatchling", "hatch-vcs"]
build-backend = "hatchling.build"

[project]
name = "my-app"
dynamic = ["version"]

[tool.hatch.version]
source = "vcs"
""",
    ) as project:
        yield project


@pytest.fixture
def new_project_write(temp_dir):
    with create_project(
        temp_dir,
        """\
[build-system]
requires = ["hatchling", "hatch-vcs"]
build-backend = "hatchling.build"

[project]
name = "my-app"
dynamic = ["version"]

[tool.hatch.version]
source = "vcs"

[tool.hatch.build.targets.wheel.hooks.vcs]
version-file = "my_app/_version.py"
""",
    ) as project:
        yield project


@pytest.fixture
def new_project_fallback(temp_dir):
    with create_project(
        temp_dir,
        """\
[build-system]
requires = ["hatchling", "hatch-vcs"]
build-backend = "hatchling.build"

[project]
name = "my-app"
dynamic = ["version"]

[tool.hatch.version]
source = "vcs"
fallback-version = "7.8.9"
""",
        setup_vcs=False,
    ) as project:
        yield project



# SPDX-FileCopyrightText: 2022-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import pytest

from hatch_vcs.version_source import VCSVersionSource


class TestTagPattern:
    def test_correct(self, new_project_basic):
        config = {'tag-pattern': '.+'}
        version_source = VCSVersionSource(new_project_basic, config)

        assert version_source.config_tag_pattern == '.+'

    def test_not_string(self, new_project_basic):
        config = {'tag-pattern': 9000}
        version_source = VCSVersionSource(new_project_basic, config)

        with pytest.raises(TypeError, match='option `tag-pattern` must be a string'):
            _ = version_source.config_tag_pattern


class TestFallbackVersion:
    def test_correct(self, new_project_basic):
        config = {'fallback-version': '0.0.1'}
        version_source = VCSVersionSource(new_project_basic, config)

        assert version_source.config_fallback_version == '0.0.1'

    def test_not_string(self, new_project_basic):
        config = {'fallback-version': 9000}
        version_source = VCSVersionSource(new_project_basic, config)

        with pytest.raises(TypeError, match='option `fallback-version` must be a string'):
            _ = version_source.config_fallback_version


class TestRawOptions:
    def test_correct(self, new_project_basic):
        config = {'raw-options': {'normalize': False}}
        version_source = VCSVersionSource(new_project_basic, config)

        assert version_source.config_raw_options == {'normalize': False}

    def test_not_table(self, new_project_basic):
        config = {'raw-options': 9000}
        version_source = VCSVersionSource(new_project_basic, config)

        with pytest.raises(TypeError, match='option `raw-options` must be a table'):
            _ = version_source.config_raw_options


def test_coverage(new_project_basic):
    version_source = VCSVersionSource(new_project_basic, {})

    assert version_source.config_tag_pattern is version_source.config_tag_pattern
    assert version_source.config_fallback_version is version_source.config_fallback_version
    assert version_source.config_raw_options is version_source.config_raw_options



# SPDX-FileCopyrightText: 2022-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import os
import sys
import zipfile

import pytest

from .utils import build_project, read_file


def test_basic(new_project_basic):
    build_project('-t', 'wheel')

    build_dir = os.path.join(new_project_basic, 'dist')
    assert os.path.isdir(build_dir)

    artifacts = os.listdir(build_dir)
    assert len(artifacts) == 1
    wheel_file = artifacts[0]

    assert wheel_file == 'my_app-1.2.3-py2.py3-none-any.whl'

    extraction_directory = os.path.join(os.path.dirname(new_project_basic), '_archive')
    os.mkdir(extraction_directory)

    with zipfile.ZipFile(os.path.join(build_dir, wheel_file), 'r') as zip_archive:
        zip_archive.extractall(extraction_directory)

    metadata_directory = os.path.join(extraction_directory, 'my_app-1.2.3.dist-info')
    assert os.path.isdir(metadata_directory)

    package_directory = os.path.join(extraction_directory, 'my_app')
    assert os.path.isdir(package_directory)
    assert len(os.listdir(package_directory)) == 4

    assert os.path.isfile(os.path.join(package_directory, '__init__.py'))
    assert os.path.isfile(os.path.join(package_directory, 'foo.py'))
    assert os.path.isfile(os.path.join(package_directory, 'bar.py'))
    assert os.path.isfile(os.path.join(package_directory, 'baz.py'))


def test_write(new_project_write):
    build_project('-t', 'wheel')

    build_dir = os.path.join(new_project_write, 'dist')
    assert os.path.isdir(build_dir)

    artifacts = os.listdir(build_dir)
    assert len(artifacts) == 1
    wheel_file = artifacts[0]

    assert wheel_file == 'my_app-1.2.3-py2.py3-none-any.whl'

    extraction_directory = os.path.join(os.path.dirname(new_project_write), '_archive')
    os.mkdir(extraction_directory)

    with zipfile.ZipFile(os.path.join(build_dir, wheel_file), 'r') as zip_archive:
        zip_archive.extractall(extraction_directory)

    metadata_directory = os.path.join(extraction_directory, 'my_app-1.2.3.dist-info')
    assert os.path.isdir(metadata_directory)

    package_directory = os.path.join(extraction_directory, 'my_app')
    assert os.path.isdir(package_directory)
    assert len(os.listdir(package_directory)) == 5

    assert os.path.isfile(os.path.join(package_directory, '__init__.py'))
    assert os.path.isfile(os.path.join(package_directory, 'foo.py'))
    assert os.path.isfile(os.path.join(package_directory, 'bar.py'))
    assert os.path.isfile(os.path.join(package_directory, 'baz.py'))

    version_file = os.path.join(package_directory, '_version.py')
    assert os.path.isfile(version_file)

    lines = read_file(version_file).splitlines()
    assert lines[3] == "version = '1.2.3'"


@pytest.mark.skipif(sys.version_info[0] == 2, reason='Depends on fix in 6.4.0 which is Python 3-only')
def test_fallback(new_project_fallback):
    build_project('-t', 'wheel')

    build_dir = os.path.join(new_project_fallback, 'dist')
    assert os.path.isdir(build_dir)

    artifacts = os.listdir(build_dir)
    assert len(artifacts) == 1
    wheel_file = artifacts[0]

    assert wheel_file == 'my_app-7.8.9-py2.py3-none-any.whl'

    extraction_directory = os.path.join(os.path.dirname(new_project_fallback), '_archive')
    os.mkdir(extraction_directory)

    with zipfile.ZipFile(os.path.join(build_dir, wheel_file), 'r') as zip_archive:
        zip_archive.extractall(extraction_directory)

    metadata_directory = os.path.join(extraction_directory, 'my_app-7.8.9.dist-info')
    assert os.path.isdir(metadata_directory)

    package_directory = os.path.join(extraction_directory, 'my_app')
    assert os.path.isdir(package_directory)
    assert len(os.listdir(package_directory)) == 4

    assert os.path.isfile(os.path.join(package_directory, '__init__.py'))
    assert os.path.isfile(os.path.join(package_directory, 'foo.py'))
    assert os.path.isfile(os.path.join(package_directory, 'bar.py'))
    assert os.path.isfile(os.path.join(package_directory, 'baz.py'))



# SPDX-FileCopyrightText: 2022-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import os

import pytest

from hatch_vcs.build_hook import VCSBuildHook


class TestVersionFile:
    def test_correct(self, new_project_basic):
        config = {'version-file': 'foo/_version.py'}
        build_dir = os.path.join(new_project_basic, 'dist')
        build_hook = VCSBuildHook(new_project_basic, config, None, None, build_dir, 'wheel')

        assert build_hook.config_version_file == 'foo/_version.py'

    def test_not_string(self, new_project_basic):
        config = {'version-file': 9000}
        build_dir = os.path.join(new_project_basic, 'dist')
        build_hook = VCSBuildHook(new_project_basic, config, None, None, build_dir, 'wheel')

        with pytest.raises(TypeError, match='Option `version-file` for build hook `vcs` must be a string'):
            _ = build_hook.config_version_file

    def test_missing(self, new_project_basic):
        config = {}
        build_dir = os.path.join(new_project_basic, 'dist')
        build_hook = VCSBuildHook(new_project_basic, config, None, None, build_dir, 'wheel')

        with pytest.raises(ValueError, match='Option `version-file` for build hook `vcs` is required'):
            _ = build_hook.config_version_file


class TestTemplate:
    def test_correct(self, new_project_basic):
        config = {'template': '__version__ = {version!r}'}
        build_dir = os.path.join(new_project_basic, 'dist')
        build_hook = VCSBuildHook(new_project_basic, config, None, None, build_dir, 'wheel')

        assert build_hook.config_template == '__version__ = {version!r}'

    def test_not_string(self, new_project_basic):
        config = {'template': 9000}
        build_dir = os.path.join(new_project_basic, 'dist')
        build_hook = VCSBuildHook(new_project_basic, config, None, None, build_dir, 'wheel')

        with pytest.raises(TypeError, match='Option `template` for build hook `vcs` must be a string'):
            _ = build_hook.config_template


def test_coverage(new_project_basic):
    config = {'version-file': 'foo/_version.py'}
    build_dir = os.path.join(new_project_basic, 'dist')
    build_hook = VCSBuildHook(new_project_basic, config, None, None, build_dir, 'wheel')

    assert build_hook.config_version_file is build_hook.config_version_file
    assert build_hook.config_template is build_hook.config_template
